#!/usr/bin/env python

"""
Solve (Del + ik) x (Del + ik) E = lambda E in a unit cell applying
periodic boundary conditions on the faces. The domain is a cuboid
intersected by 3 conductors with rectangular cross sections. The
conductors are orthogonal to each other.

"""
import math, cmath, time, sys, Tkinter
try:
    import csparse, cvector, csuperlu
except:
    print "Go and install ellipt2d from http://ellipt2d.sourceforge.net"
    sys.exit(1)
try:
    import innerfaces, brick, cluster, ctkvis
except:
    print "Some Curly modules are missing!"
    sys.exit(2)

# convenience proc's
def isInside(dindex1, dindex2, half_width1, half_width2):
    if abs(dindex1)<half_width1 and abs(dindex2)<half_width2:
        return 1
    else: return 0

# constants
PI= math.pi
TWOPI= 2.*PI

class curly:

    """
    Solve (Del + ik) x (Del + ik) E = lambda E
    
    members:
    --------
    Ns = (nx,ny,nz) no of bricks in (x,y,z) direction
    Ks = (kx,ky,kz) wavenumbers
    Ls = (Lx, Ly, lz) Domain sizes
    holeSizes = (hx,hy,hz) with Nx, Ny, Nz a multiple of hx/2, hy/2, hz/2,
    resp.

    """

    def __init__(self):
        
        " c'tor"
        
        self.Ns = (18,18,18)
        self.Ks = (0.,0.,0.)
        self.Ls = (TWOPI,TWOPI,TWOPI)
        self.holeSizes = (2*self.Ns[0]/6,2*self.Ns[1]/6,2*self.Ns[2]/6,)

        # Delta tuning
        self.Delta = 5./6.

        # solution vector
        self.v = None

    def __del__(self):

        " d'tor "
        try: del self._cluster
        except: pass
        try: del self._amat
        except: pass
        try: del self._bmat
        except: pass
        try: del self.v
        except: pass
        

    def assemble(self):

        " assemble the finite element matrices and apply BCs "
        
        Lx, Ly, Lz = self.Ls
        kx, ky, kz = self.Ks
        Nx, Ny, Nz = self.Ns
        dxyz = (Lx/Nx, Ly/Ny, Lz/Nz)
        pxyz = (kx*Lx, ky*Ly, kz*Lz)
        i0, j0, k0 = Nx/2-0.5, Ny/2-0.5, Nz/2-0.5
        di, dj, dk = self.holeSizes[0]/2,self.holeSizes[1]/2,self.holeSizes[2]/2
        
        self._amat = csparse.csparse({})
        self._bmat = csparse.csparse({})
        
        self._cluster = cluster.cluster()
        for k in range(Nz):
            for j in range(Ny):
                for i in range(Nx):
                    if isInside(i-i0, k-k0, di, dk) or \
                       isInside(i-i0,j-j0,di,dj) or \
                       isInside(j-j0,k-k0,dj,dk):
                        pass
##                        self._cluster.glueFG((i,j,k),
##                                            self._amat,
##                                            self._bmat,
##                                            dxyz, self.Ns, pxyz,
##                                            Delta=self.Delta,
##                                             G=((-0.5+0j, 0j, 0j),(-0.5+0j, 0j,),(-0.5+0j,)))
                    else:
                        self._cluster.glue((i,j,k),
                                            self._amat,
                                            self._bmat,
                                            dxyz, self.Ns, pxyz,
                                            Delta=self.Delta)

        innerEdges = innerfaces.getInnerEdges(self._cluster)

        # apply zero Bcs conditions on inner edges
        # set large number on diagonal
        size = (dxyz[0]*dxyz[1]*dxyz[2])**(1./3.) 
        for e in innerEdges:
            self._amat[(e,e)] = 1.43975e8 * size

        self.Nedges = self._cluster.Nedges        

    def solve(self, v0=None, lam0=0.3+0j, tol=1.e-6, niter=10, solver='superlu'):

        """
        Solve eigenvalue problem
        lam0 = initial guess
        tol = tolerance
        niter = max number of iterations

        return self.lambda and self.v (the solution)

        """

        if v0==None:
            # random initial eigensolution guess
            v0 = cvector.random(self.Nedges, .0, 1.+0j)

        import time
        tic=time.time()
        if solver!='distsolve':
            [self.Lambda, self.v, residue, iter] = csuperlu.eigen(
                self._amat, self._bmat, lam0, v0, tol, niter)
        else:
            import distsolve
            [self.Lambda, self.v, residue, iter] = distsolve.eigen(
                self._amat, self._bmat, lam0, v0, tol, niter,
                host = "http://petrel002.pppl.gov", port=8080)
        toc = time.time()
        print ' took ', iter, ' iterations. Time in secs %10.2f'%(toc-tic)

        vmod, varg = self.v.AbsArg()
        vmax, vmin = max(vmod), min(vmod)
        self.v = self.v/vmax


    def show(self, width=960, height=740, the=0.4156*PI, phi=0.2345*PI,
             cx=None, cy=None):

        " plot solution "

        self.root = Tkinter.Tk()

        frame = Tkinter.Frame(self.root)
        frame.pack()
        canvas = Tkinter.Canvas(bg="black", width=width, height=height)
        canvas.pack()

        p = ctkvis.ctkvis(self.Ns, width, height)
        if cx==None:
            p.CX = 0.6*p.w
        else: p.CX = cx*p.w
        if cy==None:
            p.CY = 0.95*p.h
        else: p.CY = cy*p.h
        #p.SCALE *= 1.4
        p.the = the
        p.phi = phi
        #p.edges(canvas, self._cluster, self.v)
        p.arrows(canvas, self._cluster, self.v)
        p.title(canvas, "%10.6f"% (self.Lambda.real))
        #p.exposedEdges(canvas, cluster=self._cluster)
        p.axes(canvas=canvas, cluster=self._cluster, Ks=self.Ks)
        p.colorbar(canvas)

        self.root.mainloop()
            

########################################################################

if __name__ == '__main__':

    from math import *
    from time import time
    import sys
    """
    c.Ks = (0.2,0.,0.3)
    0.3516
    0.386198
    0.408066

    """

    c = curly()
    n = 12
    c.Ns = (n,n,n)
    # dimensions
    c.Ls = (TWOPI, TWOPI, TWOPI)
    # k vector
    c.Ks = (0.5, 0., 0.0)
    # holes
    d = n/3 # n/5
    c.holeSizes = (d, d, d)
    # FE-FD mixing parameter
    c.Delta = 0.9 # 5./6.
    tic = time()
    print 'assemble...'
    c.assemble()
    toc = time()
    print 'time to assemble %f secs ' % (toc-tic)
    print '%d edges' % c.Nedges
    print 'solve...'
    if len(sys.argv)>1:
        c.solve(lam0=float(sys.argv[1]), tol=1.e-6, niter=20, solver='distsolve')
    else:
        c.solve(lam0=0.3, tol=1.e-6, niter=20)

    c.show()
    #c.show(the=0., phi=0., cx=0.05, cy=0.95)

    # write AVS/Express file
    from ucd import ucd
    u = ucd("/tmp/curly.inp")
    u.nodeVector(c._cluster, c.v, c.Ks, \
                 'E: lambda = %10.5e k=(%5.3f %5.3f %5.3f)' % \
                 (c.Lambda.real, c.Ks[0], c.Ks[1], c.Ks[2]))

##    # manipulate the data
##    import edgefilter
##    curl_of_v = edgefilter.curl(c._cluster, c.v, c.Ks)
##    poynting = edgefilter.cross(c._cluster, c.v.conjugate(), curl_of_v)

##    c.v = w
##    c.show()
    
##    u2 = ucd("/tmp/curly2.inp")
##    u2.nodeVector(c._cluster, w, c.Ks, \
##                 'CurlE: lambda = %10.5e k=(%5.3f %5.3f %5.3f)' % \
##                 (c.Lambda.real, c.Ks[0], c.Ks[1], c.Ks[2]))
    
    

